from __future__ import annotations

import re
import numpy as np
import numpy.typing as npt
import pandas as pd

from functools import partial
from logging import getLogger
from typing import NamedTuple, Callable, Literal
from matplotlib import colormaps
from matplotlib import colors
from statsmodels.nonparametric.smoothers_lowess import lowess


from ._context import Context, Data, FigureRuntime, FigureSettings
from ._handler import CommandHandler, Token, CommandResult
from ._numberunit import NumberUnit, Bound, parse_range, NumberUnitRange
from ._misc import nderivative

from ._parser import CommandParser

QUALITATIVE_CMAPS = {'Pastel1':9,'Pastel2':8,'Paired':12,'Accent':8,'Dark2':8,'Set1':9,'Set2':8,'Set3':12,'tab10':10,'tab20':20,'tab20b':20,'tab20c':20}

class Labels(NamedTuple):
    title: str | None
    xlabel: str | None
    ylabel: str | None

class ColorMap(NamedTuple):
    cmap: colors.Colormap
    num: int | None

class Args_Plot(NamedTuple):
    plot: ColKind | None
    figure: FigureSettings
    labels: Labels | None
    suptitle: str
    xlimits: NumberUnitRange | None
    ylimits: NumberUnitRange | None
    vshift: float
    colorby: str | None
    colormap: ColorMap
    alpha: float
    linewidth: float
    linestyle: str | tuple[Literal[0], tuple[int,...]]
    legend: bool
    show: bool
    grid: str|None

class ColKind(NamedTuple):
    xcol: str | None
    ycol: str
    xop:list[Callable[[npt.NDArray], npt.NDArray]]
    yop:list[Callable[[npt.NDArray, npt.NDArray], npt.NDArray]]
    xdef: str
    ydef: str

def derivative(y:npt.NDArray, x:npt.NDArray, d:int=1) -> npt.NDArray:
    return nderivative(y, d, x) # type: ignore

def weight(y:npt.NDArray, x:npt.NDArray, d:int=0) -> npt.NDArray:
    if d <= 0: return y
    return y * x ** d

def weight_noavg(y:npt.NDArray, x:npt.NDArray, d:int=0) -> npt.NDArray:
    if d <= 0: return y
    ym = y.mean()
    return (y - ym) * x ** d + ym

def weight_one(y:npt.NDArray, x:npt.NDArray, d:int=0) -> npt.NDArray:
    if d <= 0: return y
    return (y - 1) * x ** d + 1 # type: ignore

def smooth(y:npt.NDArray, x:npt.NDArray, d:int=1) -> npt.NDArray:
    return lowess(y, x, d/len(x), it=0, is_sorted=False, return_sorted=False)

# XY_CLN_KIND = re.compile(r"^([\w.+*(,)-]+)\:([\w.+*(,)-]+)$")
XY_CLN_KIND = re.compile(r"^(?:([\w.+*(,)-]+)\:)?([\w.+*(,)-]+)$")
def parse_column(p:str) -> ColKind:
    m = XY_CLN_KIND.match(p)
    if m is None: raise ValueError(f"Invalid plot type: {p}")

    x_specs, y_specs = m.groups()
    x_ops, y_ops = [],[]

    if x_specs is not None:
        x_spec = x_specs.split(".")
        for xs in reversed(x_spec[:-1]):
            match xs:
                case "r": op = lambda x: np.real(x)
                case "i": op = lambda x: np.imag(x)
                case "a": op = lambda x: np.abs(x)
                case "p": op = lambda x: np.unwrap(np.angle(x))
                case _: raise ValueError(f"Unknown x operation: {xs}")
            x_ops.append(op)
        xcol = x_spec[-1]
    else:
        xcol = None

    y_spec = y_specs.split(".")
    # Check that there is an index if there are operations. If there is not, error
    if len(y_spec) > 1 and x_specs is None:
        raise ValueError("Y operations require an X column to operate on.")

    for ys in reversed(y_spec[:-1]):
        factor = int(ys[1:]) if len(ys) > 1 else 1
        match ys[0]:
            case "r": op = lambda y,_: np.real(y)
            case "i": op = lambda y,_: np.imag(y)
            case "a": op = lambda y,_: np.abs(y)
            case "p": op = lambda y,_: np.unwrap(np.angle(y))
            case "s": op = lambda y,x,f=factor: partial(smooth,       d=f)(y,x)
            case "d": op = lambda y,x,f=factor: partial(derivative,   d=f)(y,x)
            case "k": op = lambda y,x,f=factor: partial(weight,       d=f)(y,x)
            case "w": op = lambda y,x,f=factor: partial(weight_one,   d=f)(y,x)
            case "W": op = lambda y,x,f=factor: partial(weight_noavg, d=f)(y,x)
            case _: raise ValueError(f"Unknown x operation: {ys}")
        y_ops.append(op)
    ycol = y_spec[-1]

    return ColKind(xcol, ycol, x_ops, y_ops, x_specs, y_specs)

def get_column_(col:ColKind, data:Data) -> tuple[npt.NDArray, npt.NDArray]:
    domain = data._get_col_domain(col.ycol)
    x = data.get_col_(col.xcol, domain=domain)
    for op in col.xop:
        x = op(x)

    y = data.get_col_(col.ycol, domain=domain)
    for op in col.yop:
        y = op(y,x)

    return x,y

def get_series(col:ColKind, data:Data) -> pd.Series:
    x,y = get_column_(col, data)
    return pd.Series(y, x)


FIGURE_SUBPLOT = re.compile(r"^(\d+)(?::(\d+)\.(\d+))?(;)?$")
class Plot(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Plot:
        parser = CommandParser("plot", description="Plots the calculated data.")
        parser.add_argument("data", nargs="?")
        parser.add_argument("--xlabel")
        parser.add_argument("--ylabel")
        parser.add_argument("--title")
        parser.add_argument("--suptitle")
        parser.add_argument("--xlim", nargs=2)
        parser.add_argument("--ylim", nargs=2)
        parser.add_argument("--vshift", type=float, default=0)
        parser.add_argument("--colorby")
        parser.add_argument("--figure", default=None)
        parser.add_argument("--color", nargs="+")
        parser.add_argument("--alpha", type=float, default=1.0)
        parser.add_argument("--show", action="store_true")
        parser.add_argument("--legend", action="store_true")
        parser.add_argument("--gridx", action="store_true")
        parser.add_argument("--gridy", action="store_true")
        parser.add_argument("--grid", action="store_true")

        lwidth = parser.add_mutually_exclusive_group()
        lwidth.add_argument("--linewidth", type=float, default=1.0)
        lwidth.add_argument("--xxthick", action="store_const", const=8.0, dest="linewidth")
        lwidth.add_argument("--xthick", action="store_const", const=4.0, dest="linewidth")
        lwidth.add_argument("--thick", action="store_const", const=2.0, dest="linewidth")
        
        lwidth.add_argument("--thin", action="store_const", const=0.5, dest="linewidth")
        lwidth.add_argument("--xthin", action="store_const", const=0.25, dest="linewidth")
        lwidth.add_argument("--xxthin", action="store_const", const=0.125, dest="linewidth")

        lstyle = parser.add_mutually_exclusive_group()
        lstyle.add_argument("--linestyle", default="solid")
        lstyle.add_argument("--solid", action="store_const", const="solid", dest="linestyle")
        lstyle.add_argument("--dotted", action="store_const", const="dotted", dest="linestyle")
        lstyle.add_argument("--dashed", action="store_const", const="dashed", dest="linestyle")
        lstyle.add_argument("--dashdot", action="store_const", const="dashdot", dest="linestyle")
        
        args = parser.parse(tokens)

        plot = parse_column(args.data) if args.data is not None else None
        
        # Select color and colorby
        match args.color:
            case None:
                cmap = colormaps["tab10"]
                cmapcount = QUALITATIVE_CMAPS.get("tab10")
            case [str(_cmapname)] if _cmapname in colormaps:
                cmap = colormaps[_cmapname]
                cmapcount = QUALITATIVE_CMAPS.get(_cmapname)
            case [str(_color)]:
                # Double up the color to have a constant colormap
                cmap = colors.LinearSegmentedColormap.from_list("customcmap", [_color,_color])
                cmapcount = None
            case list(_colorlist):
                cmap = colors.LinearSegmentedColormap.from_list('customcmap', _colorlist)
                cmapcount = None
            case _:
                raise RuntimeError()
        colormap = ColorMap(cmap, cmapcount)
        
        # Labels
        if any(l is not None for l in (args.xlabel, args.ylabel, args.title)):
            labels = Labels(
                args.title,
                args.xlabel,
                args.ylabel,
            )
        else: labels = None

        # Figure, show
        show = args.show
        if args.figure is None:
            figsettings = FigureSettings(-1, (1,1))
            figure = figsettings
            context.figures.impl_figsettings.append(figsettings)
        else:
            m = FIGURE_SUBPLOT.match(args.figure)
            if m is None:
                raise ValueError(f"Invalid figure specification: \"{args.figure}\"")
            match m.groups()[0:3]:
                case str(_f),None,None:
                    fignum,a,b = int(_f), 1, 1
                case str(_f),None,None:
                    fignum,a,b = int(_f), 1, 1
                case str(_f),str(_a),str(_b):
                    fignum,a,b = int(_f),int(_a),int(_b)
                case u:
                    raise ValueError(f"Invalid figure specification: \"{u}\"")
            
            if not show and m.group(4) == ";": show = True
            
            figure = FigureSettings(fignum, (a,b))

            if fignum not in context.figures.expl_figsettings:
                context.figures.expl_figsettings[fignum] = figure
            else:
                _prev_fig = context.figures.expl_figsettings[fignum]
                # Update the figure size to the maximum subplot count
                at,bt = (max(a, _prev_fig.subplot[0]),max(b, _prev_fig.subplot[1]))
                context.figures.expl_figsettings[fignum] = FigureSettings(fignum, (at,bt))

        # Line style
        # Try parsing as a list of numbers separated by dots
        try:
            style = (0, tuple(int(n) for n in str(args.linestyle).split(".")))
        except ValueError:
            style = args.linestyle
        
        match args.gridx, args.gridy, args.grid:
            case [False, False, False]: grid = None
            case [True, True, False] | [_, _, True]: grid = "xy"
            case [True, False, False]: grid = "x"
            case [False, True, False]: grid = "y"
            case _:raise RuntimeError("Boolean logic goes brrr")

        return Args_Plot(
            plot,
            figure,
            labels,
            args.suptitle,
            parse_range(*args.xlim) if args.xlim else None,
            parse_range(*args.ylim) if args.ylim else None,
            args.vshift,
            args.colorby,
            colormap,
            args.alpha,
            args.linewidth,
            style,
            args.legend,
            show,
            grid
        )

    @staticmethod
    def execute(args: Args_Plot, context: Context) -> CommandResult:
        log = getLogger("plot")

        figurenum = args.figure.figurenum
        if figurenum == -1:
            # The figure is in impl_figsettings. Get a nonallocated number.
            figurenum = context.figures.get_impl_figurenum()
            figsettings = FigureSettings(figurenum, (1,1))
            figure = FigureRuntime.new(figsettings)
            context.figures.figureruntimes[figurenum] = figure
        else:
            # The figure was explicitly allocated. Get the actual preexisting figure.
            if figurenum not in context.figures.figureruntimes:
                figsettings = context.figures.expl_figsettings[figurenum]
                figure = FigureRuntime.new(figsettings)

                context.figures.figureruntimes[figurenum] = figure
            else:
                figure = context.figures.figureruntimes[figurenum]

        fig, ax = figure.figure, figure.axes[args.figure.subplot]

        if args.plot is not None:
            # Get coloring
            if args.colorby is None:
                _colorby = np.array(range(len(context.data.data))) + len(ax._lines) + 1
            else:
                _colorby = np.array([data.meta.get(args.colorby) for data in context.data])
            qualitative = np.issubdtype(_colorby.dtype, np.integer) or np.issubdtype(_colorby.dtype, np.str_)
            
            match qualitative, args.colormap.num:
                case True, None:
                # Value is qualitative and colormap is quantitative
                    uniquevals, inverse = np.unique(_colorby, return_inverse=True)
                    ncats = len(uniquevals)
                    catcolors = np.array([args.colormap.cmap(i/(max(ncats-1,1))) for i,_ in enumerate(uniquevals)])
                    colors = catcolors[inverse]
                case True, cmaplength:
                # Value is qualitative and colormap is qualitative
                    uniquevals, inverse = np.unique(_colorby, return_inverse=True)
                    ncats = len(uniquevals)
                    catcolors = np.array([args.colormap.cmap(i%cmaplength) for i,_ in enumerate(uniquevals)])
                    colors = catcolors[inverse]
                case False, _:
                # Value is quantitative
                    m,M = min(_colorby), max(_colorby)
                    if np.isclose(m-M, 0):
                        colors = args.colormap.cmap(np.zeros_like(_colorby))
                    else:
                        colors = args.colormap.cmap((_colorby-m)/(M-m))
                    pass
                    

            for i,data in enumerate(context.data):
                x,_y = get_column_(args.plot, data)
                y = _y + i * args.vshift

                ax._lines.append((x,y))
                
                ax.axis.plot(x, y, color=colors[i],
                             linewidth=args.linewidth,
                             linestyle=args.linestyle,
                             alpha=args.alpha,
                             label=data.meta.name
                            )

        if args.suptitle is not None:
            fig.suptitle(args.suptitle)

        if args.labels is not None:
            if args.labels.xlabel is not None:
                ax.axis.set_xlabel(args.labels.xlabel)
            if args.labels.ylabel is not None:
                ax.axis.set_ylabel(args.labels.ylabel)
            if args.labels.title is not None:
                ax.axis.set_title(args.labels.title)


        # Update the stored limit definitions, if not `..`
        if args.xlimits is not None:
            _al, _au = ax.xlimits
            match args.xlimits.lower:
                case NumberUnit(_l, _, _) if _l == -np.inf:...
                case _:
                    _al = args.xlimits.lower
            match args.xlimits.upper:
                case NumberUnit(_u, _, _) if _u == np.inf:...
                case _:
                    _au = args.xlimits.upper
            ax.xlimits = (_al, _au)
            
        if args.ylimits is not None:
            _al, _au = ax.ylimits
            match args.ylimits.lower:
                case NumberUnit(_l, _, _) if _l == -np.inf:...
                case _:
                    _al = args.ylimits.lower
            match args.ylimits.upper:
                case NumberUnit(_u, _, _) if _u == np.inf:...
                case _:
                    _au = args.ylimits.upper
            ax.ylimits = (_al, _au)
        
        # Act on the limit definitions, if the limit is not None or inf / -inf
        match ax.xlimits[0]:
            case Bound.EXTERNAL:
                _xlow = min([np.min(x) for x,_ in ax._lines])
            case Bound.INTERNAL:
                _xlow = max([np.min(x) for x,_ in ax._lines])
            case NumberUnit(_l,_,_) if _l != -np.inf:
                _xlow = _l
            case _:
                _xlow = None
        if _xlow is not None:
            ax.axis.set_xlim(left = _xlow)

        match ax.xlimits[1]:
            case Bound.EXTERNAL:
                _xhig = max([np.max(x) for x,_ in ax._lines])
            case Bound.INTERNAL:
                _xhig = min([np.max(x) for x,_ in ax._lines])
            case NumberUnit(_l,_,_) if _l != np.inf:
                _xhig = _l
            case _:
                _xhig = None
        if _xhig is not None:
            ax.axis.set_xlim(right = _xhig)

        _xlow,_xhig = ax.axis.get_xlim()
        
        match ax.ylimits[0]:
            case Bound.EXTERNAL:
                _ylow = min([np.min(y[(x>=_xlow)&(x<=_xhig)]) for x,y in ax._lines])
            case Bound.INTERNAL:
                _ylow = max([np.min(y[(x>=_xlow)&(x<=_xhig)]) for x,y in ax._lines])
            case NumberUnit(_l,_,_) if _l != -np.inf:
                _ylow = _l
            case _:
                _ylow = None

        match ax.ylimits[1]:
            case Bound.EXTERNAL:
                _yhig = max([np.max(y[(x>=_xlow)&(x<=_xhig)]) for x,y in ax._lines])
            case Bound.INTERNAL:
                _yhig = min([np.max(y[(x>=_xlow)&(x<=_xhig)]) for x,y in ax._lines])
            case NumberUnit(_l,_,_) if _l != np.inf:
                _yhig = _l
            case _:
                _yhig = None
        
        match ax.ylimits[0], ax.ylimits[1]:
            case Bound(), Bound():
                # Expand both limits by 5%
                _d = _yhig-_ylow # type: ignore
                _yhig = _yhig + _d * 0.05 # type: ignore
                _ylow = _ylow - _d * 0.05 # type: ignore
            case Bound(), _:
                # Expand lower limit by 5% relative to current upper ylim
                _,_u = ax.axis.get_ylim()
                _d = _u-_ylow # type: ignore
                _ylow = _ylow - _d * 0.05 # type: ignore
            case _, Bound():
                # Expand upper limit by 5% relative to current lower ylim
                _l,_ = ax.axis.get_ylim()
                _d = _yhig-_l # type: ignore
                _yhig = _yhig + _d * 0.05 # type: ignore

        if _ylow is not None:
            ax.axis.set_ylim(bottom = _ylow)
        if _yhig is not None:
            ax.axis.set_ylim(top = _yhig)
        
        if args.legend:
            ax.axis.legend()
        
        if args.grid is not None:
            if "x" in args.grid:ax.axis.grid(True, axis="x")
            if "y" in args.grid:ax.axis.grid(True, axis="y")
        

        if args.show:
            figure.show()

        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError



from typing import Iterable

ORDINAL = {"1":"st", "2":"nd", "3":"rd"}
CHR = "0123456789+-=()"
SUP = "⁰¹²³⁴⁵⁶⁷⁸⁹⁺⁻⁼⁽⁾"
SUB = "₀₁₂₃₄₅₆₇₈₉₊₋₌₍₎"

def th(n:int) -> str:
    """Return the ordinal representation of a number as a string.
    >>> th(1), th(2), th(3), th(4), th(11), th(21)
    ('1st', '2nd', '3rd', '4th', '11th', '21st')
    """
    return f"{n}{ORDINAL.get(str(n)[-1], "th")}"

def plu(n:int, w:str, p:str|None=None) -> str:
    """Return the plural form of a word based on the given number.
    If `p` is provided, it is used as the plural form; otherwise, 's' is appended to `w`.
    >>> plu(1, "cat")
    'cat'
    >>> plu(2, "cat")
    'cats'
    >>> plu(2, "person", "people")
    'people'
    """
    return w if n == 1 else (p or w + "s")

def are(n:int) -> str:
    """Return 'is' if n == 1 else 'are'."""
    return "is" if n == 1 else "are"

def sup(v:int|str) -> str:
    """Convert a string to its superscript representation. Non-supported characters are unchanged.
    >>> sup("32")
    '³²
    """
    return "".join(SUP[CHR.index(c)] if c in CHR else c for c in str(v))

def sub(v:int|str) -> str:
    """Convert a string to its subscript representation. Non-supported characters are unchanged.
    >>> sub("H2O")
    'H₂O'
    """
    return "".join(SUB[CHR.index(c)] if c in CHR else c for c in str(v))

def exp(v:float, ndig:int=1) -> str:
    """Convert a float to a string in scientific notation with superscript exponent.
    The number of digits in the mantissa can be specified with `ndig`.
    >>> exp(12345)
    '1.2x10⁴'
    >>> exp(0.0012345, 2)
    '1.23x10⁻³'
    """
    preformatted = format(v, f"0.{ndig}e")
    if "+" in preformatted or "-" in preformatted:
        val,e = preformatted.split("e")
        while len(e) > 2 and e[1] == "0":
            e = e[0] + e[2:]
        if e == "+0": return f"{val}"
        return f"{val}{sup(e)}"
    else:
        return preformatted

def pol(p:Iterable[float], sep:str=" ") -> str:
    """Convert polynomial coefficients to string.
    Coefficients are in the order of x^0, x^1, ..., x^n
    >>> pol([1,2,3])
    '1x⁰ 2x¹ 3x²'
    >>> pol([1,2,3], sep=" + ")
    '1x⁰ + 2x¹ + 3x²'
    """
    return sep.join(f"{exp(a)}x{sup(e)}" for e,a in enumerate(p))



from __future__ import annotations

import numpy as np
import numpy.typing as npt
import pandas as pd
import re

from typing import NamedTuple
from logging import getLogger
from dataclasses import dataclass
from scipy.interpolate import interp1d
from pathlib import Path


from ._context import Context, MetaData
from ._handler import CommandHandler, Token, CommandResult
from ._numberunit import NumberUnit, parse_range, NumberUnitRange, actualize_range

from ._parser import CommandParser

from .plot import ColKind, parse_column, get_column_


def save_df_to_dat(filename:str|Path, df:pd.DataFrame,*, index:bool=False, head:str|None=None, mode:str="wt"):
    colwidth = 16
    decformat = np.vectorize(lambda x,cw=colwidth:format(x, "<10.6f").ljust(cw))
    sdecformat = np.vectorize(lambda x,cw=colwidth:format(x, "<+10.6f").ljust(cw))
    sciformat = np.vectorize(lambda x,cw=colwidth:format(x, "<10.6e").ljust(cw))
    ssciformat = np.vectorize(lambda x,cw=colwidth:format(x, "<+10.6e").ljust(cw))

    def formatcol(x:npt.NDArray) -> npt.NDArray[np.str_]:
        sci = ((x != 0)&(np.abs(np.log(np.abs(x + np.finfo(x.dtype).eps))) >= 4)).any()
        sgn = (x < 0).any()
        match bool(sci),bool(sgn):
            case False, False:
                fmt = decformat
            case False, True :
                fmt = sdecformat
            case True , False:
                fmt = sciformat
            case True , True :
                fmt = ssciformat
        return fmt(x)

    if index:
        headers = [str(df.index.name).ljust(colwidth)] + [str(c).ljust(colwidth) for c in df.columns]
        columns = [formatcol(df.index.to_numpy())] + [formatcol(df.loc[:, col].to_numpy()) for col in df.columns]
    else:
        headers = [str(c).ljust(colwidth) for c in df.columns]
        columns = [formatcol(df.loc[:, col].to_numpy()) for col in df.columns]

    stacked = np.column_stack(columns)

    with open(filename, mode) as fw:
        if head:
            fw.write(head + "\n")
        fw.write(f"#L {"".join(headers).rstrip()}\n")
        fw.writelines(f"   {"".join(l).rstrip()}\n" for l in stacked)

def get_var_headers(meta:MetaData) -> str:
    r = re.compile(r"\.h(\d+)\.(\d+)")
    _head:list[list[str]] = []
    for varname, val in meta.vars.items():
        m = r.match(varname)
        if m is None: continue

        _line,_pos = m.groups()
        line = int(_line) - 1
        pos = int(_pos) - 1

        while len(_head) <= line:
            _head.append([])
        while len(_head[line]) <= pos:
            _head[line].append("")
        _head[line][pos] = str(val)
    
    head = [" ".join(line) for line in _head]

    if ".st" in meta:
        head.append(f"#V signal {meta.get(".st")}")
    if ".rt" in meta:
        head.append(f"#V reference {meta.get(".rt")}")
    if "E0" in meta:
        head.append(f"#V E0 {meta.get("E0")}")
    if "rE0" in meta:
        head.append(f"#V refE0 {meta.get("rE0")}")

    for vname, v in meta.vars.items():
        if vname.startswith("."):
            continue
        head.append(f"#V {vname} {v}")
    
    return "\n".join(head)


@dataclass(slots=True, frozen=True)
class Mode:
    filename: str

@dataclass(slots=True, frozen=True)
class Each(Mode):
    columns: list[ColKind]

@dataclass(slots=True, frozen=True)
class Batch(Mode):
    column: ColKind

@dataclass(slots=True, frozen=True)
class Aligned(Mode):
    align: NumberUnitRange
    column: ColKind

class Args_Save(NamedTuple):
    mode: Mode


class Save(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Save:
        parser = CommandParser("save", description="Saves the data to a file.")
        parser.add_argument("filename")

        subparsers = parser.add_subparsers(dest="mode")
        aligned = subparsers.add_parser("aligned")
        aligned.add_argument("column")
        aligned.add_argument("--align", "-a", nargs=3, required=True)

        batch = subparsers.add_parser("batch")
        batch.add_argument("column")

        columns = subparsers.add_parser("columns")
        columns.add_argument("columns", nargs="+")
        
        args = parser.parse(tokens)

        match args.mode:
            case "batch":
                col = parse_column(args.column)
                if col.xcol is None:
                    raise ValueError("No x column specified.")
                mode = Batch(args.filename, col)
                
            case "aligned":
                col = parse_column(args.column)
                if col.xcol is None:
                    raise ValueError("No x column specified.")

                if col.xop:
                    raise ValueError("The output x column cannot be modified.")
                
                bounds = parse_range(*args.align)
                if bounds.lower.value == -np.inf:
                    raise ValueError("Alignment range cannot be ..")
                if bounds.upper.value == np.inf:
                    raise ValueError("Alignment range cannot be ..")
                    
                mode = Aligned(args.filename, bounds, col)
            case "columns":
                cols = [parse_column(c) for c in args.columns]
                mode = Each(args.filename, cols)

        
        return Args_Save(mode)

    @staticmethod
    def execute(args: Args_Save, context: Context) -> CommandResult:
        log = getLogger("save")
        
        match args.mode:
            case Batch(fname, column):
                outfile = context.paths.outputdir / fname
                cols:list[pd.Series] = []
                for data in context.data:
                    x,y = get_column_(column, data)
                    X = pd.Series(x, name=f"{column.xdef}_{data.meta.name}")
                    cols.append(X)
                    Y = pd.Series(y, name=f"{column.ydef}_{data.meta.name}")
                    cols.append(Y)
                df = pd.concat(cols, axis=1)
                save_df_to_dat(outfile, df, index=False)
                log.info(f"Saved data batch to {outfile} as {column.xdef}:{column.ydef}.")

            case Aligned(fname, align, column):
                outfile = context.paths.outputdir / fname
                if column.xcol is None:
                    raise RuntimeError("No x column specified.")
                domain = context.data.data[0]._get_col_domain(column.xcol)
                if domain != align.domain:
                    raise RuntimeError("Invalid range selection: range does not match given abscissa.")
                unit = context.data.data[0].datums[domain].cols[column.xcol].unit
                bounds = actualize_range(align, [data.get_col_(column.xcol) for data in context.data], unit)
                
                match bounds.inter:
                    case NumberUnit(i, _, _):
                        newx = np.arange(bounds.lower.value, bounds.upper.value, i)
                    case int(n):
                        newx = np.linspace(bounds.lower.value, bounds.upper.value, n)
                    case _:
                        raise RuntimeError("Unknown error. #23794618")
                
                out:list[pd.Series] = []
                x = pd.Index(newx, name = column.xcol)
                for data in context.data:
                    X,Y = get_column_(column, data)
                    y = interp1d(X,Y,"cubic")(newx)
                    out.append(pd.Series(y, x, name=data.meta.name))
                df = pd.concat(out, axis=1)
                save_df_to_dat(outfile, df, index=True)
                
                log.info(f"Saved aligned data to {outfile} as {column.xdef}:{column.ydef}.")


            case Each(fname, columns):
                outnames:dict[str,int] = {}
                for data in context.data:
                    filename = fname
                    for m in reversed([*re.finditer(r"\{([^{:}]*)(?::([^{:}]*))?\}", fname)]):
                        var,fmt = m.groups()
                        val = data.meta.get(var)
                        val = "" if val is None else val
                        val = format(val, fmt) if fmt is not None else str(val)
                        filename = filename[:m.start()] + str(val) + filename[m.end():]
                    outfile = context.paths.outputdir / filename
                    if outfile.name in outnames:
                        n = outnames[outfile.name]
                        newname = outfile.name.removesuffix(outfile.suffix) + f"_{n}" + outfile.suffix
                        outfile = context.paths.outputdir / newname
                        outnames[outfile.name] = n + 1
                    else:
                        outnames[outfile.name] = 1

                    cols = [pd.Series(get_column_(c, data)[1], name=c.ydef) for c in columns]
                    df = pd.concat(cols, axis=1)
                    head = get_var_headers(data.meta)
                    save_df_to_dat(outfile, df, index=False, head=head)
                    
                    log.info(f"{data.meta.name}: Saved file to {outfile} with columns {" ".join(c.ydef for c in columns)}.")
    
        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError



from __future__ import annotations

import numpy as np
import numpy.typing as npt

from enum import Enum
from logging import getLogger
from typing import NamedTuple

from ._context import Context, AxisType, Column, DataColType, FourierType, Domain, range_to_index
from ._handler import CommandHandler, Token, CommandResult
from ._numberunit import NumberUnit, parse_range, NumberUnitRange, actualize_range
from ._parser import CommandParser

class Apodizer(Enum):
    HANN = "hann"
    RECTANGLE = "rect"
    BARTLETT = "bart"
    WELCH = "welc"
    SINE = "sine"
    EXPONENTIAL = "expn"
    GAUSS = "gauss"

class Method(Enum):
    DFT = "dft"
    FINUFT = "finuft"
    FFT = "fft"


class Args_Fourier(NamedTuple):
    bounds: NumberUnitRange
    fbounds: NumberUnitRange
    kweight: float
    apodizer: tuple[Apodizer, float | None]
    wwidth: float
    method: Method

def get_window(_r:npt.NDArray[np.floating], apodizer:Apodizer, parameter: float|None=None, symm:bool=True) -> npt.NDArray[np.floating]:
    r0 = np.clip(_r, 0, None)
    if symm:
        r = np.clip(np.abs(_r), 0, 1)
    else:
        r = np.clip(_r, 0, 1)

    match apodizer:
        case Apodizer.HANN:
            return (np.cos(np.pi * r) + 1) / 2
        case Apodizer.RECTANGLE:
            return (r < 1)*1.0
            # return (r <= 0.5)*1.0
        case Apodizer.BARTLETT:
            return 1 - r
        case Apodizer.WELCH:
            return 1 - r * r
        case Apodizer.SINE:
            return np.cos(np.pi * r / 2)
        case Apodizer.EXPONENTIAL:
            assert parameter is not None
            return np.exp(-r0 * parameter)
        case Apodizer.GAUSS:
            assert parameter is not None
            return np.exp(-r0 * r0 * parameter)
    

def get_flattop_window(
    x: npt.NDArray[np.floating],
    width: np.floating | float,
    apodizer: Apodizer,
    parameter: float | None = None,
    symmetry:int = 0,
    shift: tuple[float,float] | float | None = None
) -> npt.NDArray[np.floating]:
    if width == 0:
        return np.ones_like(x)
    
    width = np.min([(np.max(x) - np.min(x))/2, np.float64(width)])

    if symmetry > 0:
        if shift is None: shift = 0
        _r = np.array((np.zeros_like(x), x + width - x.max() - shift))
    elif symmetry < 0:
        if shift is None: shift = 0
        _r = np.array((np.zeros_like(x), x.min() + shift - x + width))
    else:
        if shift is None: shift = 0,0
        _r = np.array((np.zeros_like(x), x + width - x.max() - shift[1], x.min() + shift[0] - x + width)) # type: ignore
    r = np.max(_r,axis=0,)/ width

    return get_window(r, apodizer, parameter)

def fourier(
    x: npt.NDArray[np.floating],
    y: npt.NDArray[np.floating | np.complexfloating],
    r: npt.NDArray[np.floating],
) -> npt.NDArray[np.complexfloating]:
    dx = np.diff(x)

    FT = 2 * np.outer(x, r)
    FTl, FTr = FT[:-1, :], FT[1:, :]

    yl = y[:-1] * dx
    yr = y[1:] * dx

    R: npt.NDArray[np.floating] = yl.dot(np.cos(FTl)) + yr.dot(np.cos(FTr))
    I: npt.NDArray[np.floating] = yl.dot(np.sin(FTl)) + yr.dot(np.sin(FTr))

    return (R + 1.0j * I) / np.sqrt(2 * np.pi)

def bfourier(
    r: npt.NDArray[np.floating],
    y: npt.NDArray[np.floating] | npt.NDArray[np.complexfloating],
    x: npt.NDArray[np.floating],
) -> npt.NDArray[np.complexfloating]:
    return fourier(r, y.conj(), x).conj()

def finuft(
    x: npt.NDArray[np.floating],
    y: npt.NDArray[np.complexfloating],
    r: npt.NDArray[np.floating],
) -> npt.NDArray[np.complexfloating]:
    from finufft import nufft1d3
    return nufft1d3(x, y, 2*r, eps=1e-9)

def fft(
    x: npt.NDArray[np.floating],
    y: npt.NDArray[np.floating],
    *,
    correct_r: bool=False
) -> tuple[npt.NDArray[np.floating], npt.NDArray[np.complexfloating]]:
    
    # Ensure proper x spacing, below a tolerance of 1e-5
    diff = np.diff(x)
    if np.diff(diff).max()/diff.mean() > 1e-5: raise RuntimeError("X spacing is not consistent.")

    # Automatic 2 zerofilling
    n = int(np.exp2(np.ceil(np.log2(len(x)))+2))

    r = np.fft.rfftfreq(n, diff.mean())
    if correct_r:
        r = r * np.pi
    f = np.fft.rfft(y,n)
    return r,f

class Fourier(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Fourier:
        parser = CommandParser("fourier", description="Performs the discrete Fourier transform.")
        parser.add_argument("range", nargs=2, help="The transformation range.")
        parser.add_argument("outrange", nargs=2, help="The maximum R value and spacing of the DFT.")
        parser.add_argument("--kweight", "-k", type=float, default=0, help="The k-weight to scale the signal by.")
        parser.add_argument("--width", "-w", type=float, default=np.inf, help="The ramp width of the apodizer window.")
        parser.add_argument("--apodizer", "-a", default="hann", help="The window function to apodize the data with.")
        parser.add_argument("--method",default="dft",choices=["dft", "finuft", "fft"])

        args = parser.parse(tokens)

        range = parse_range(*args.range)
        outrn = parse_range("0A", *args.outrange)

        match args.apodizer[:3]:
            case "han":
                apd = (Apodizer.HANN, None)
            case "sin":
                apd = (Apodizer.SINE, None)
            case "rec":
                apd = (Apodizer.RECTANGLE, None)
            case "bar" | "tri":
                apd = (Apodizer.BARTLETT, None)
            case "wel":
                apd = (Apodizer.WELCH, None)
            case "exp":
                _, p = args.apodizer.split(":", maxsplit=1)
                apd = (Apodizer.EXPONENTIAL, float(p))
            case "gau":
                _, p = args.apodizer.split(":", maxsplit=1)
                apd = (Apodizer.GAUSS, float(p))
        
        match args.method:
            case "dft":
                met = Method.DFT
            case "finuft":
                met = Method.FINUFT
            case "fft":
                met = Method.FFT

        return Args_Fourier(range, outrn, args.kweight, apd, args.width, met)

    @staticmethod
    def execute(args: Args_Fourier, context: Context) -> CommandResult:
        log = getLogger("fourier")

        if args.fbounds.domain != Domain.FOURIER:
            raise RuntimeError("The fourier transform requires valid fourier units.")
        
        match args.fbounds.inter:
            case int(n):
                R = np.linspace(args.fbounds.lower.value, args.fbounds.upper.value, n)
            case NumberUnit(v, _, "A"):
                R = np.arange(args.fbounds.lower.value, args.fbounds.upper.value, v)
            case _: raise RuntimeError("Invalid R axis specification.")
        interval = R[1] - R[0]

        domain = args.bounds.domain or Domain.RECIPROCAL
        if domain != Domain.RECIPROCAL:
            raise RuntimeError("Cannot fit postedge to a different domain.")
        _axes = [data.get_col_(data.datums[domain].default_axis) for data in context.data] # type: ignore
        range = actualize_range(args.bounds, _axes, "eV")

        for data in context.data:

            idx = range_to_index(data, range)

            X = data.get_col_("k")
            Y = data.get_col_("x")
            x,_y = X[idx], Y[idx]

            w = get_flattop_window(x, args.wwidth, *args.apodizer) # type: ignore

            # Check if the data has mean 1. If so, the background was not removed,
            # and is estimated as 1
            kw = np.ones(x.shape, dtype=np.float64) if args.kweight == 0 else x ** args.kweight # type: ignore
            kw: npt.NDArray[np.floating]
            y = w * _y * kw

            # Approximate 1/Nyquist frequency
            _min_dR = 2 * np.diff(x).mean()

            log.debug(f"{data.meta.name}: Performing fourier transform")

            if _min_dR > interval:
                log.warning(
                    f"The required frequency ({1/interval:0.1f}) is smaller than the average Nyquist frequency ({1/_min_dR:0.1f}). The minimum interval is {_min_dR:0.4f}"
                )

            match args.method:
                case Method.DFT:
                    f = fourier(x, y, R) # type: ignore
                case Method.FINUFT:
                    f = finuft(x, y+0j, R) # type: ignore
                case Method.FFT:
                    _R,_f = fft(x, y, correct_r=True)
                    selidx = (_R>=args.fbounds.lower.value)&(_R<=args.fbounds.upper.value)
                    R,f = _R[selidx], _f[selidx]

            W = np.zeros_like(X)
            W[idx] = w

            data.add_col("win", W, Column(None, None, DataColType.WINDOW), Domain.RECIPROCAL)

            data.add_col("R", R, Column("A", None, AxisType.DISTANCE), Domain.FOURIER)
            data.datums[Domain.FOURIER].set_default_axis("R")
            
            data.add_col("f", f, Column(None, None, FourierType.COMPLEX), Domain.FOURIER)

        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError



import re
import numpy as np
import numpy.typing as npt

from typing import overload

from ._numberunit import parse_nu, SI_multiplier

def attempt_parse_number(s: str) -> str | int | float:
    if s.isdigit():
        return int(s)
    try:
        return float(s)
    except ValueError:
        return s
    
_ELEMENT_EDGE = r"(\w+)\.(\w+)(?:([+-]?\d+(?:\.\d+)?)(?:([kMGTPEZYRQmunpfazyrq])?(eV))?)?"
ELEMENT_EDGE = re.compile(_ELEMENT_EDGE)


def parse_edgeenergy(val: str) -> float:
    try:
        nu = parse_nu(val)
        if nu.unit not in ("eV", None):
            raise ValueError(f'Invalid energy unit: "{nu.unit}"')

        return nu.value
    except ValueError:
        m = ELEMENT_EDGE.match(val)

        if m is None: 
            raise ValueError(f'Could not parse "{val}" as edge energy.')
        
        element, edge, value, mult, _ = m.groups()

        shift = float(value) * SI_multiplier(mult) if value is not None else 0

        from xraydb import xray_edge

        energy = xray_edge(element, edge, True)
        return energy + shift  # type: ignore


# eV to k conversion

KCONST = 0.26246842396479844746511289042048746313978437879180
SCONST = 0.51231672231618445138819837616078387648071784401809  # sqrt(KCONST)
KICNST = 3.8099821109685837311852822702410407881602562961373  # 1 / KCONST


@overload
def E_to_k(v: float, E0: np.floating | float) -> float: ...
@overload
def E_to_k(v: np.floating, E0: np.floating | float) -> np.floating: ...
@overload
def E_to_k(v: npt.NDArray[np.floating], E0: np.floating | float) -> npt.NDArray[np.floating]: ...

def E_to_k(v: float | np.floating | npt.NDArray[np.floating], E0: np.floating | float):
    return SCONST * np.sqrt((v - E0))


@overload
def k_to_E(v: float, E0: np.floating | float) -> float: ...
@overload
def k_to_E(v: np.floating, E0: np.floating | float) -> np.floating: ...
@overload
def k_to_E(v: npt.NDArray, E0: np.floating | float) -> npt.NDArray: ...
def k_to_E(v, E0: np.floating | float):
    return v**2 * KICNST + E0


# sk = signed k


@overload
def E_to_sk(v: float, E0: np.floating | float) -> float: ...
@overload
def E_to_sk(v: np.floating, E0: np.floating | float) -> np.floating: ...
@overload
def E_to_sk(v: npt.NDArray, E0: np.floating | float) -> npt.NDArray: ...
def E_to_sk(v, E0: np.floating | float):
    return SCONST * np.sqrt((np.abs(v - E0))) * np.sign(v - E0)


@overload
def sk_to_E(v: float, E0: np.floating | float) -> float: ...
@overload
def sk_to_E(v: np.floating, E0: np.floating | float) -> np.floating: ...
@overload
def sk_to_E(v: npt.NDArray, E0: np.floating | float) -> npt.NDArray: ...
def sk_to_E(v, E0: np.floating | float):
    return v**2 * KICNST * np.sign(v) + E0


from typing import SupportsIndex, Sequence
import numpy._core.numeric as _nx

def derivative2(
    f: npt.ArrayLike,
    *varargs: npt.ArrayLike,
    axis: None | SupportsIndex | Sequence[SupportsIndex] = None,
) -> npt.NDArray:
    
    # Modified from np.gradient to calculate second order derivative instead of first
    
    f = np.asanyarray(f)
    N = f.ndim  # number of dimensions

    if axis is None:
        axes = tuple(range(N))
    else:
        axes = _nx.normalize_axis_tuple(axis, N)

    len_axes = len(axes)
    n = len(varargs)
    
    if n == 0:
        # no spacing argument - use 1 in all axes
        dx = [1.0] * len_axes
    elif n == 1 and np.ndim(varargs[0]) == 0:
        # single scalar for all axes
        dx = varargs * len_axes
    elif n == len_axes:
        # scalar or 1d array for each axis
        dx = list(varargs)
        for i, distances in enumerate(dx):
            distances = np.asanyarray(distances)
            if distances.ndim == 0:
                continue
            elif distances.ndim != 1:
                raise ValueError("distances must be either scalars or 1d")
            if len(distances) != f.shape[axes[i]]:
                raise ValueError("when 1d, distances must match "
                                 "the length of the corresponding dimension")
            if np.issubdtype(distances.dtype, np.integer):
                # Convert numpy integer types to float64 to avoid modular
                # arithmetic in np.diff(distances).
                distances = distances.astype(np.float64)
            diffx = np.diff(distances)
            # if distances are constant reduce to the scalar case
            # since it brings a consistent speedup
            if (diffx == diffx[0]).all():
                diffx = diffx[0]
            dx[i] = diffx
    else:
        raise TypeError("invalid number of arguments")
    
    slice1 = [slice(None)]*N
    slice2 = [slice(None)]*N
    slice3 = [slice(None)]*N
    slice4 = [slice(None)]*N

    otype = f.dtype
    if np.issubdtype(otype, np.inexact):
        pass
    else:
        # All other types convert to floating point.
        # First check if f is a numpy integer type; if so, convert f to float64
        # to avoid modular arithmetic when computing the changes in f.
        if np.issubdtype(otype, np.integer):
            f = f.astype(np.float64)
        otype = np.float64
    
    
    for axis, ax_dx in zip(axes, dx):
        if f.shape[axis] < 3:
            raise ValueError(
                "Shape of array too small to calculate a numerical gradient, "
                "at least 3 elements are required.")
        # result allocation
        out = np.empty_like(f, dtype=otype)

        outvals = []

        slice1[axis] = slice(1, -1)
        slice2[axis] = slice(None, -2)
        slice3[axis] = slice(1, -1)
        slice4[axis] = slice(2, None)

        # spacing for the current axis
        uniform_spacing = np.ndim(ax_dx) == 0
        if uniform_spacing:
            out[tuple(slice1)] = (f[tuple(slice2)] - 2*f[tuple(slice3)] + f[tuple(slice4)]) / (ax_dx**2)
        
        else:
            dx1 = ax_dx[:-1]
            dx2 = ax_dx[1:]

            A = 1 / (dx1 * (dx1 + dx2))
            B = -1 / (dx1 * dx2)
            C = 1 / (dx2 * (dx1 + dx2))
            # fix the shape for broadcasting
            shape = np.ones(N, dtype=int)
            shape[axis] = -1
            A.shape = B.shape = C.shape = shape
            
            out[tuple(slice1)] = (A * f[tuple(slice2)] + B * f[tuple(slice3)] + C * f[tuple(slice4)]) * 2.0

        ### DONE -----------------------------------------------------------------------------------------------------------------------

        # Edges
        slice1[axis] = 0
        slice2[axis] = 1
        slice3[axis] = 2
        out[tuple(slice1)] = out[tuple(slice2)] + (out[tuple(slice2)] - out[tuple(slice3)]) / ax_dx[1] * ax_dx[0]
        slice1[axis] = -1
        slice2[axis] = -2
        slice3[axis] = -3
        out[tuple(slice1)] = out[tuple(slice2)] + (out[tuple(slice2)] - out[tuple(slice3)]) / ax_dx[-2] * ax_dx[-1]

        outvals.append(out)

        # reset the slice object in this dimension to ":"
        slice1[axis] = slice(None)
        slice2[axis] = slice(None)
        slice3[axis] = slice(None)
        slice4[axis] = slice(None)

    if len_axes == 1:
        return outvals[0]
    return tuple(outvals)

def nderivative(f: npt.ArrayLike,d:int,
    *varargs: npt.ArrayLike,
    axis: None | SupportsIndex | Sequence[SupportsIndex] = None,):

    f = np.asanyarray(f)
    N = f.ndim  # number of dimensions

    if axis is None:
        axes = tuple(range(N))
    else:
        axes = _nx.normalize_axis_tuple(axis, N)
    
    if d <= 0:
        return f
    elif d == 1:
        return np.gradient(f, *varargs, axis=axes)
    elif d == 2:
        return derivative2(f, *varargs, axis=axes)
    
    if len(axes) == 1:
        for i in range(d//2):
            f = derivative2(f, *varargs, axis=axes)
        if d%2!=0:
            f = np.gradient(f, *varargs, axis=axes)
        return f
    else:
        fs = list(derivative2(f, *varargs, axis=axes))
        for i in range(d//2 -1):
            fs = [derivative2(_f, _x, axis=_axis) for _f,_x,_axis in zip(fs, varargs, axes)]
        if d%2!=0:
            fs = [np.gradient(_f, _x, axis=_axis) for _f,_x,_axis in zip(fs, varargs, axes)]
        return tuple(fs)



from abc import ABC, abstractmethod
from shlex import split
from typing import NamedTuple

from ._parser import Token
from ._context import Context, CommandResult


class AbstractCommandHandler(ABC):

    @staticmethod
    @abstractmethod
    def tokenize(lines: list[tuple[int, str]]) -> list[Token]: ...

    @staticmethod
    @abstractmethod
    def parse(tokens: list[Token], context: Context) -> NamedTuple: ...

    @staticmethod
    @abstractmethod
    def execute(args: NamedTuple, context: Context) -> CommandResult: ...

    @staticmethod
    @abstractmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult: ...


class CommandHandler(AbstractCommandHandler):

    @staticmethod
    def tokenize(lines: list[tuple[int, str]]) -> list[Token]:
        return [
            Token(token, lineno, line.index(token))
            for lineno, line in lines
            for token in split(line)
        ]



from __future__ import annotations

import numpy as np

from logging import getLogger
from typing import NamedTuple
from statsmodels.nonparametric.smoothers_lowess import lowess

from ._context import Context, Domain, range_to_index, DataColType, Column
from ._handler import CommandHandler, Token, CommandResult
from ._numberunit import NumberUnit, parse_range, NumberUnitRange, actualize_range
from ._parser import CommandParser


class Args_Cut(NamedTuple):
    bounds: NumberUnitRange


class Cut(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Cut:
        parser = CommandParser("cut", description="Cuts the data.")
        parser.add_argument("range", nargs=2)

        args = parser.parse(tokens)

        range = parse_range(*args.range)

        return Args_Cut(range)

    @staticmethod
    def execute(args: Args_Cut, context: Context) -> CommandResult:
        log = getLogger("cut")

        domain = args.bounds.domain or Domain.RECIPROCAL
        _axes = [data.get_col_(data.datums[domain].default_axis) for data in context.data] # type: ignore
        range = actualize_range(args.bounds, _axes, "eV")
        
        for data in context.data:
            idx = range_to_index(data, range)
            data.datums[domain].df = data.datums[domain].df.loc[idx, :]
                
            log.debug(f"{data.meta.name}: Cut data in the range {range.lower.value:0.3f}{range.lower.unit} ~ {range.upper.value:0.3f}{range.upper.unit}") # type: ignore

        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError


class Args_Smooth(NamedTuple):
    data: str
    xaxis: str
    window: int


class Smooth(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Smooth:
        parser = CommandParser("smooth", description="Smooths the data.")
        parser.add_argument("data", nargs="?", default="a")
        parser.add_argument("xaxis", nargs="?", default="E")
        parser.add_argument("--window", "-w", default=5, type=int, help="The smoothing window.")

        args = parser.parse(tokens)

        return Args_Smooth(args.data, args.xaxis, args.window)

    @staticmethod
    def execute(args: Args_Smooth, context: Context) -> CommandResult:
        log = getLogger("smooth")

        for data in context.data:
            d = data.get_xy(args.xaxis, args.data)
            _frac = args.window / len(d)

            x,y = d.index.to_numpy(), d.to_numpy()
            sy = lowess(y,x, _frac, 0, return_sorted=False)
            data.mod_col(args.data, sy)
            log.debug(f"{data.meta.name}: smoothed column {args.data} with lowess.")
                    
        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError


class Args_Rebin(NamedTuple):
    region: NumberUnitRange

class Rebin(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Rebin:
        parser = CommandParser("rebin", description="Rebins data.")
        parser.add_argument("region", nargs=2)
        parser.add_argument("interval", nargs="?")
        parser.add_argument("--number", "-n", nargs="?", type=int)

        args = parser.parse(tokens)
        # TODO: range must be inclusive
        # Ensure all points are satisfied
        if args.number is not None:
            region = parse_range(*args.region, number=args.number)
        else:
            if args.interval is None:
                raise ValueError("Interval or number of points must be specified.")
            region = parse_range(*args.region, args.interval)

        return Args_Rebin(region)

    @staticmethod
    def execute(args: Args_Rebin, context: Context) -> CommandResult:
        log = getLogger("rebin")

        domain = args.region.domain
        if domain is None:
            raise RuntimeError("Rebin domain is unspecified.")

        if args.region.lower.value == -np.inf or args.region.upper.value == np.inf:
            raise RuntimeError("Rebin range must be specified. Cannot be \"..\".")
        
        match args.region:
            case NumberUnitRange(_, _, None):
                raise RuntimeError("No interval specified for rebin.")

            case NumberUnitRange(_, _, NumberUnit(_,-1|1,"eV")): axis, unit = "e", "eV"
            case NumberUnitRange(_, _, NumberUnit(_,0,"eV")): axis, unit = "E", "eV"
            case NumberUnitRange(_, _, NumberUnit(_,_,"k")): axis, unit = "k", "k"
            case NumberUnitRange(_, _, NumberUnit(_,_,"A")): axis, unit = "R", "A"

            case NumberUnitRange(NumberUnit(_,-1|1,"eV"), _, _): axis, unit = "e", "eV"
            case NumberUnitRange(NumberUnit(_,0,"eV"), _, _): axis, unit = "E", "eV"
            case NumberUnitRange(NumberUnit(_,_,"k"), _, _): axis, unit = "k", "k"
            case NumberUnitRange(NumberUnit(_,_,"A"), _, _): axis, unit = "R", "A"
            
            case NumberUnitRange(_, NumberUnit(_,-1|1,"eV"), _): axis, unit = "e", "eV"
            case NumberUnitRange(_, NumberUnit(_,0,"eV"), _): axis, unit = "E", "eV"
            case NumberUnitRange(_, NumberUnit(_,_,"k"), _): axis, unit = "k", "k"
            case NumberUnitRange(_, NumberUnit(_,_,"A"), _): axis, unit = "R", "A"

            case _:raise RuntimeError("No unit specified for rebin.")
        
        _axes = [data.get_col_(axis) for data in context.data] # type: ignore
        range = actualize_range(args.region, _axes, unit)
        
        match range.inter:
            case NumberUnit(i, _, _):
                newx = np.arange(range.lower.value, range.upper.value, i)
            case int(n):
                newx = np.linspace(range.lower.value, range.upper.value, n)
            case _: raise RuntimeError("Unknown error. #23794608")
        i = newx[1]-newx[0]
        bins = np.concat([[newx[0]-i/2], newx[:]+i/2])
        
        
        for data in context.data:
            X = data.get_col_(axis)
            dig = np.digitize(X, bins)
            df = data.datums[domain]
            df.df = df.df.groupby(dig).mean().loc[1:len(newx),:]
            df.df.loc[:,axis] = newx[df.df.index-1]
        
        # TODO: log
        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError



class Args_Normalize(NamedTuple):
    var: str | None
    val: float

class Normalize(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Normalize:
        parser = CommandParser("normalize", description="Normalizes XAS signal to calculate XANES and EXAFS data.")
        parser.add_argument("var", default="J0")
        parser.add_argument("--value", "-v")
        args = parser.parse(tokens)

        if args.value is not None:
            args.var = None
        return Args_Normalize(args.var, args.value)

    @staticmethod
    def execute(args: Args_Normalize, context: Context) -> CommandResult:
        log = getLogger("normalize")

        for data in context.data:
            norm = data.meta.get(args.var) if args.var is not None else args.val

            a = data.get_col_("a")

            data.add_col("mu", a / norm, Column(None, None, DataColType.MU), Domain.RECIPROCAL)
            data.add_col("x", a / norm - 1, Column(None, None, DataColType.CHI), Domain.RECIPROCAL)

        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError



from __future__ import annotations

import numpy as np

from enum import Enum
from typing import NamedTuple
from logging import getLogger


from ._context import Context, AxisType, DataColType, Column, Domain, range_to_index
from ._format import pol
from ._handler import CommandHandler, Token, CommandResult
from ._numberunit import NumberUnit, parse_range, NumberUnitRange, actualize_range

from ._parser import CommandParser


class RemovalOperation(Enum):
    DIVIDE = "/"
    SUBTRACT = "-"


class Args_PostEdge(NamedTuple):
    bounds: NumberUnitRange
    action: RemovalOperation
    fitaxis: AxisType
    degree: int


class PostEdge(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_PostEdge:
        parser = CommandParser(
            "postedge", description="Removes the background postedge contribution."
        )
        parser.add_argument("range", nargs=2, help="The polynomial fit range.")

        groupd = parser.add_mutually_exclusive_group()
        groupd.add_argument(
            "--constant",
            "-C",
            dest="degree",
            action="store_const",
            const=0,
            help="Models the postedge as a constant polynomial.",
        )
        groupd.add_argument(
            "--linear",
            "-l",
            dest="degree",
            action="store_const",
            const=1,
            help="Models the postedge as a linear polynomial.",
        )
        groupd.add_argument(
            "--quadratic",
            "-q",
            dest="degree",
            action="store_const",
            const=2,
            help="Models the postedge as a quadratic polynomial.",
        )
        groupd.add_argument(
            "--cubic",
            "-c",
            dest="degree",
            action="store_const",
            const=3,
            help="Models the postedge as a cubic polynomial.",
        )
        groupd.add_argument(
            "--polynomial",
            "-p",
            dest="degree",
            type=int,
            help="Models the postedge as a polynomial trend, and removes it from the data.",
        )

        groupm = parser.add_mutually_exclusive_group()
        groupm.add_argument(
            "--divide", "-d", dest="method", action="store_const", const="div"
        )
        groupm.add_argument(
            "--subtract", "-s", dest="method", action="store_const", const="sub"
        )

        groupa = parser.add_mutually_exclusive_group()
        groupa.add_argument(
            "--energy", "-e", dest="fitaxis", action="store_const", const="e"
        )
        groupa.add_argument(
            "--wavevector", "-k", dest="fitaxis", action="store_const", const="k"
        )

        args = parser.parse(tokens)

        range = parse_range(*args.range)
        if range.domain != Domain.RECIPROCAL:
            raise ValueError("Invalid fit domain: the postedge can only be calculated in energy or wavevector.")


        # Match bounds and fitaxis
        match args.fitaxis:
            case "e":
                fitaxis = AxisType.RELENERGY
            case "k":
                fitaxis = AxisType.KVECTOR
            case None:
                match range.lower:
                    case NumberUnit(_, _, "eV"):
                        fitaxis = AxisType.RELENERGY
                    case NumberUnit(_, _, "k"):
                        fitaxis = AxisType.KVECTOR
                    case _:
                        match range.upper:
                            case NumberUnit(_, _, "eV"):
                                fitaxis = AxisType.RELENERGY
                            case NumberUnit(_, _, "k"):
                                fitaxis = AxisType.KVECTOR
                            case _:
                                raise ValueError("Neither bound has units specified. Specify the fit axis explicitly.")

        match args.method:
            case "div":
                action = RemovalOperation.DIVIDE
            case "sub":
                action = RemovalOperation.SUBTRACT
            case _:
                raise RuntimeError("Invalid removal operation: 458314")
            
        

        return Args_PostEdge(range, action, fitaxis, args.degree)

    @staticmethod
    def execute(args: Args_PostEdge, context: Context) -> CommandResult:
        log = getLogger("postedge")

        domain = args.bounds.domain or Domain.RECIPROCAL
        if domain != Domain.RECIPROCAL:
            raise RuntimeError("Cannot fit postedge to a different domain.")
        
        _axes = [data.get_col_(data.datums[domain].default_axis) for data in context.data] # type: ignore
        range = actualize_range(args.bounds, _axes, "eV")

        for data in context.data:
            if "postedge" in data.meta.run:
                raise RuntimeError(f"Postedge was already calculated for {data.meta.name}")
            idx = range_to_index(data, range)
            log.debug(f"{data.meta.name}: Fitting postedge of order {args.degree} in the region {range.lower.value:0.3f}{range.lower.unit} ~ {range.upper.value:0.3f}{range.upper.unit}") # type: ignore

            X = data.get_col_(coltype=args.fitaxis)
            Y = data.get_col_("a")
            x, y = X[idx], Y[idx]
            poly = np.poly1d(np.polyfit(x, y, args.degree)) # type: ignore
            P = poly(X)

            log.debug(f"{data.meta.name}: postedge = {pol(poly.coef)}")
            
            data.meta.run["postedge"] = (poly, args.fitaxis, args.action)
            J0 = poly(0)
            data.meta.vars["J0"] = J0
            data.add_col("post", P, Column(None, None, DataColType.POSTEDGE), Domain.RECIPROCAL)

            match args.action:
                case RemovalOperation.SUBTRACT:
                    Y1 = (Y - P) + J0
                    data.mod_col("a", Y1)
                    # data.add_col("mu", Y1+1, Column(None, None, DataColType.MU), Domain.RECIPROCAL)
                    # data.add_col("x", Y1, Column(None, None, DataColType.CHI), Domain.RECIPROCAL)

                case RemovalOperation.DIVIDE:
                    Y1 = Y / P * J0
                    data.mod_col("a", Y1)
                    # data.add_col("mu", Y1, Column(None, None, DataColType.MU), Domain.RECIPROCAL)
                    # data.add_col("x", Y1-1, Column(None, None, DataColType.CHI), Domain.RECIPROCAL)

        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError



from __future__ import annotations

import os
import pandas as pd
import numpy as np

from io import StringIO
from logging import getLogger
from pathlib import Path
from typing import NamedTuple, Sequence
from enum import Enum

from ._exceptions import FileParsingException, ArgumentException
from ._context import AxisType, Context, Data, MetaData, SignalType, Column, Domain, DataColType
from ._handler import CommandHandler, Token, CommandResult
from ._misc import attempt_parse_number
from ._parser import CommandParser, InputFileParsingException


def _read_file_m1(file: Path) -> tuple[CommandResult, pd.DataFrame, dict[str, str | int | float]]:
    # Expect the data file to start with a "#"
    filecontent = file.read_text()
    lines = filecontent.splitlines()

    if filecontent[0] == "#":
        # raise FileParsingException("Invalid file specification. Missing header.")

        # Gets the index of the first line that does not start with #
        id_firstdataline = (
            i for i, line in enumerate(lines) if not line.lstrip().startswith("#")
        ).__next__()

        metadatalines = [line.split() for line in lines[: id_firstdataline - 1]]
        headers = lines[id_firstdataline - 1].split()[1:]
    else:
        id_firstdataline = 0
        metadatalines = []
        headers = None

    
    # Check if comma -> ,\s+
    firstdataline = lines[id_firstdataline]
    if "," in firstdataline:
        if "." in firstdataline:
            # Assume , is column separator and . is decimal separator
            sep,engine = r",\s+","python"
    else:
        sep,engine = r"\s+",None

    try:
        data = pd.read_csv(
            StringIO("\n".join(lines[id_firstdataline:])), sep=sep, header=None, engine=engine,
        )
    except pd.errors.ParserError:
        raise FileParsingException("File parsing error.")

    if headers is not None:
        datalen, headlen = len(data.columns), len(headers)
        if datalen < headlen:
            headers = headers[:datalen]
            warn = f"Header and data length mismatch: {headlen} {datalen}. Header was cut to match."
        elif datalen > headlen:
            headers.extend(f"col{i+1}" for i in range(headlen, datalen))
            warn = f"Header and data length mismatch: {headlen} {datalen}. Header was extended to match."
        else:
            warn = None
        data.columns = headers
    else:
        datalen = len(data.columns)
        data.columns = [f"col{i+1}" for i in range(datalen)]
        warn = None

    # --- Parse metadata -------------------------------------------------------
    metavars: dict[str, str | int | float] = {}

    # Add file header metadata. The split is done at every space.
    metavars.update(
        {
            f".h{ln}.{pn}": attempt_parse_number(item)
            for ln, line in enumerate(metadatalines, 1)
            for pn, item in enumerate(line, 1)
        }
    )

    # If a line starts with #U then what follows is a varname and a value
    for line in metadatalines:
        if not line:
            continue
        if line[0] == "#U" and len(line) >= 3:
            vname, vval = line[1], attempt_parse_number(line[2])
            if vname.startswith("."):
                raise NameError("Variables cannot start with dot.")
            metavars[vname] = vval

    return CommandResult(True, warning=warn), data, metavars


_file_read_methods = [_read_file_m1]


def _get_column_index(_columns: Sequence[str], column: str | int) -> int:
    if isinstance(column, int):
        return column
    if column.isdigit():
        return int(column) - 1
    if column in _columns:
        return _columns.index(column)
    raise IndexError(f"Column {column} not found.")


def _get_column_range(
    _columns: Sequence[str], cola: str | int, colb: str | int | None = None
) -> list[int]:
    if colb is None:
        return [_get_column_index(_columns, cola)]
    return [
        *range(_get_column_index(_columns, cola), _get_column_index(_columns, colb) + 1)
    ]


def _parse_column_range(_columns: Sequence[str], range: str) -> list[int]:
    return [
        col
        for interval in range.split(",")
        for col in _get_column_range(_columns, *interval.split(".."))
    ]

def get_filepaths(file:str, dir:Path) -> list[Path]:
    if os.path.isabs(file):
        _p = Path(file)
        _d = Path((_p.drive) + "/")
        _r = _p.relative_to(_d)
        return [*_d.glob(_r.name)]
    else:
        return [*dir.glob(file)]

def _read_file(file:Path, log_fname:str|None=None) -> tuple[CommandResult, pd.DataFrame, dict[str, str | int | float]]:
    log_fname = log_fname or file.name
    for method in _file_read_methods:
        return method(file)
    else:
        raise RuntimeError(f"Unrecognized file format: {log_fname}")

def read_file(file: Path, args: Args_FileIn) -> Data:
    """Reads the given file and extracts data and metadata.
    Attempts all methods before failing."""
    log = getLogger("filein")

    if args.directory is not None:
        _log_fname = file.relative_to(args.directory)
    else:
        _log_fname = file

    log.debug(f"Reading file: {_log_fname}")

    res, filedat, mdat = _read_file(file)

    if res.warning is not None:
        log.warning(f"{_log_fname}: {res.warning}")
    
    # Add file name metadata. The split is done at every underscore.
    mdat.update(
        {
            f".f{n}": attempt_parse_number(e)
            for n, e in enumerate(file.name.removesuffix(file.suffix).split("_"), 1)
        }
    )
    mdat[".f"] = file.name
    mdat[".fn"] = file.name.removesuffix(file.suffix)
    mdat[".fe"] = file.suffix

    # Add all the given variables from the command
    for vname, vval in args.vars.items():
        if vname.startswith("."):
            raise NameError("Variables cannot start with dot.")
        if isinstance(vval, str) and vval.startswith("."):
            # If the variable value starts with a period, check if it is a header
            # or filename var, and set the predefined value to a pretty $name
            mdat[vname] = mdat.get(vval, vval)
        else:
            mdat[vname] = vval

    name = file.name.removesuffix(file.suffix)
    signaltype = args.signaltype[0] if args.signaltype is not None else None
    refsigtype = args.refsigtype[0] if args.refsigtype is not None else None
    metadata = MetaData(signaltype, refsigtype, name, file, mdat)

    headercolumns = list(filedat.columns)
    dat = Data(metadata=metadata)

    # Import the main axis column
    xaxiscol = _get_column_index(headercolumns, args.xaxiscol)
    xcolumn = filedat.iloc[:, xaxiscol]
    match args.axis:
        case AxisType.INDEX:
            signaldomain = Domain.RECIPROCAL
        case AxisType.ENERGY:
            dat.add_col("E", xcolumn, Column("eV", False, AxisType.ENERGY), Domain.RECIPROCAL)
            signaldomain = Domain.RECIPROCAL
            dat.datums[signaldomain].set_default_axis("E")
        case AxisType.RELENERGY:
            dat.add_col("e", xcolumn, Column("eV", True, AxisType.RELENERGY), Domain.RECIPROCAL)
            signaldomain = Domain.RECIPROCAL
            dat.datums[signaldomain].set_default_axis("e")
        case AxisType.KVECTOR:
            dat.add_col("k", xcolumn, Column("k", None, AxisType.KVECTOR), Domain.RECIPROCAL)
            signaldomain = Domain.RECIPROCAL
            dat.datums[signaldomain].set_default_axis("k")
        case AxisType.DISTANCE:
            dat.add_col("R", xcolumn, Column("A", None, AxisType.DISTANCE), Domain.FOURIER)
            signaldomain = Domain.FOURIER
            dat.datums[signaldomain].set_default_axis("R")
        case AxisType.QVECTOR:
            dat.add_col("q", xcolumn, Column("q", None, AxisType.QVECTOR), Domain.RECIPROCAL)
            signaldomain = Domain.RECIPROCAL
            dat.datums[signaldomain].set_default_axis("q")
    
    # Import the intensities
    for cname, (crange, ctype) in args.columns.items():
        dat.add_col(
            cname,
            filedat.iloc[:,_parse_column_range(headercolumns, crange)].sum(axis=1),
            ctype,
            signaldomain
        )

    if args.signaltype is not None:
        stype, scols = args.signaltype
        match stype:
            case SignalType.INTENSITY:
                scolumn = dat.get_col(scols[0])
                dat.add_col("a", scolumn, Column(None, None, SignalType.INTENSITY), signaldomain)
            case SignalType.TRANSMITTANCE:
                scolumn = np.log10(dat.get_col(scols[0]) / dat.get_col(scols[1]))
                dat.add_col("a", scolumn, Column(None, None, SignalType.TRANSMITTANCE), signaldomain)
            case SignalType.FLUORESCENCE:
                scolumn = dat.get_col(scols[0]) / dat.get_col(scols[1])
                dat.add_col("a", scolumn, Column(None, None, SignalType.FLUORESCENCE), signaldomain)

    if args.refsigtype is not None:
        rtype, rcols = args.refsigtype
        match rtype:
            case SignalType.INTENSITY:
                scolumn = dat.get_col(rcols[0])
                dat.add_col("ref", scolumn, Column(None, None, SignalType.INTENSITY), signaldomain)
            case SignalType.TRANSMITTANCE:
                scolumn = np.log10(dat.get_col(rcols[0]) / dat.get_col(rcols[1]))
                dat.add_col("ref", scolumn, Column(None, None, SignalType.TRANSMITTANCE), signaldomain)
            case SignalType.FLUORESCENCE:
                scolumn = dat.get_col(rcols[0]) / dat.get_col(rcols[1])
                dat.add_col("ref", scolumn, Column(None, None, SignalType.FLUORESCENCE), signaldomain)

    # Sort the axis
    axis = dat.datums[signaldomain].default_axis
    if axis is not None:
        df = dat.datums[signaldomain].df
        df.sort_values(axis, inplace=True)
        df = df.groupby(axis, as_index=False).mean()
        dat.datums[signaldomain].df = df
    return dat


class Args_FileIn(NamedTuple):
    filepos: str
    directory: str | None

    axis: AxisType
    xaxiscol: str
    shift: float

    columns: dict[str, tuple[str, Column]]

    signaltype: tuple[SignalType, tuple[str, ...]]
    refsigtype: tuple[SignalType, tuple[str, ...]] | None

    vars: dict[str, str | int | float]


class BatchIn(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> None:
        if "batcharguments" not in context.options.other:
            context.options.other["batcharguments"] = list[list[Token]]()
        context.options.other["batcharguments"].append(tokens)

    @staticmethod
    def execute(args: None, context: Context) -> CommandResult:
        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError


class FileIn(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_FileIn:
        parser = CommandParser("filein", description="Imports the data from files.")
        parser.add_argument("filepos", help="The file to import. Can be a glob pattern.")
        parser.add_argument("--dir", help="The directory to import the files from, instead of the current workdir.", )
        parser.add_argument("--batch", "-b", action="store_true", help="Uses the batch arguments")
        parser.add_argument("--xaxiscolumn", "-X", help="Column containing the x-axis.")

        groupx = parser.add_mutually_exclusive_group()
        groupx.add_argument("--index", nargs="?", type=str, default=None, const=True, help="Sets the x axis to a numeric index.")
        groupx.add_argument("--energy", "-E", nargs="?", type=str, default=None, const=True, help="Reads the x-axis column as energy.")
        groupx.add_argument("--kvector", "-k", nargs="?", type=str, default=None, const=True, help="Reads the x-axis column as k-vector.")
        groupx.add_argument("--rdistance", "-R", nargs="?", type=str, default=None, const=True, help="Reads the x-axis column as radial distance.")
        groupx.add_argument("--qvector", "-q", nargs="?", type=str, default=None, const=True, help="Reads the x-axis column as q-vector.")
        parser.add_argument("--shift", type=float, default=0, help="Shifts the data by the specified amount.")
        parser.add_argument("--intensities","-I",nargs="+",help="Imports raw intensities, in order: I0, I1, I2, If. If not specified, determines columns from data import commands. To skip one, use '-'",)

        groupd = parser.add_mutually_exclusive_group()
        groupd.add_argument("--transmission","-t",nargs="*",help="Reads transmission data, and calculates as log(#1/#2).",)
        groupd.add_argument("--fluorescence","-f",nargs="*",help="Reads fluorescence data, and calculates as #1/#2.",)
        groupd.add_argument("--intensity", "-i", help="Reads signal, as is.")
        groupd.add_argument("--mu", "-m", help="Imports the specified column as a XANES signal.")
        groupd.add_argument("--chi", "-x", help="Imports the specified column as an EXAFS signal.")

        groupr = parser.add_mutually_exclusive_group()
        groupr.add_argument("--reftransmittance","-T",nargs="*",help="Reads transmission reference data, and calculates as log(#1/#2).",)
        groupr.add_argument("--refabsorption", "-A", help="Reads reference signal, as is.")

        parser.add_argument("--var", action="append", nargs=2)

        # Preparse the current arguments. If batch is used, reparse the arguments
        # parsing the batch arguments first, to update the namespace.
        args = parser.parse(tokens)
        if args.batch:
            if "batcharguments" not in context.options.other:
                raise ArgumentException("No batch arguments specified.")
            namespace = None
            for batcharg in context.options.other["batcharguments"]:
                batcharg: list[Token]
                namespace = parser.parse([Token("", -1, -1), *batcharg], namespace)
            args = parser.parse(tokens, namespace)

        # --- Detect the input x axis type -------------------------------------

        match args.xaxiscolumn, args.index, args.energy, args.kvector, args.rdistance, args.qvector:
            case [None, None, None, None, None, None]:
                raise InputFileParsingException("No x axis specified.")
            case [col, None, None, None, None, None]:
                xaxiscol, xaxis = col, AxisType.INDEX
            case [None, col, None, None, None, None]:
                xaxiscol, xaxis = col, AxisType.INDEX
            case [col, True, None, None, None, None]:
                xaxiscol, xaxis = col, AxisType.INDEX
            case [None, None, col, None, None, None]:
                xaxiscol, xaxis = col, AxisType.ENERGY
            case [col, None, True, None, None, None]:
                xaxiscol, xaxis = col, AxisType.ENERGY
            case [None, None, None, col, None, None]:
                xaxiscol, xaxis = col, AxisType.KVECTOR
            case [col, None, None, True, None, None]:
                xaxiscol, xaxis = col, AxisType.KVECTOR
            case [None, None, None, None, col, None]:
                xaxiscol, xaxis = col, AxisType.DISTANCE
            case [col, None, None, None, True, None]:
                xaxiscol, xaxis = col, AxisType.DISTANCE
            case [None, None, None, None, None, col]:
                xaxiscol, xaxis = col, AxisType.QVECTOR
            case [col, None, None, None, None, True]:
                xaxiscol, xaxis = col, AxisType.QVECTOR
            case _:
                raise InputFileParsingException("Invalid x axis type combination.")

        # --- Intensities, transmission, absorption, fluorescence are only valid
        # if the x axis is ENERGY. If else, use intensity ----------------------

        cols: dict[str, tuple[str, Column]] = {}

        match xaxis, args.transmission, args.fluorescence, args.intensities, args.intensity, args.mu, args.chi:
            case [AxisType.ENERGY, None, None, None, None, None, None]:
                # None of transmission, fluorescence, intensities, intensity specified.
                raise InputFileParsingException("No y axis specified.")
            case [AxisType.ENERGY, [str(cI)], None, None, None, None, None]:
                ytype = None
                cols["a"] = cI, Column(None, None, DataColType.ALPHA)
            case [AxisType.ENERGY, [str(cI0), str(cI1)], None, None, None, None, None]:
                # --transmission I0 I1
                ytype = SignalType.TRANSMITTANCE, ("I0", "I1")
                cols["I0"] = cI0, Column(None, None, SignalType.INTENSITY)
                cols["I1"] = cI1, Column(None, None, SignalType.INTENSITY)
            case [AxisType.ENERGY, None, [str(cI)], None, None, None, None]:
                ytype = SignalType.FLUORESCENCE, ("I0", "I1")
                cols["a"] = cI, Column(None, None, DataColType.ALPHA)
            case [AxisType.ENERGY, None, [str(cIf), str(cI0)], None, None, None, None]:
                # --fluorescence If I0
                ytype = SignalType.FLUORESCENCE, ("If", "I0")
                cols["I0"] = cI0, Column(None, None, SignalType.INTENSITY)
                cols["If"] = cIf, Column(None, None, SignalType.INTENSITY)
            case [AxisType.ENERGY, None, None, [str(cI)], None, None, None]:
                # --intensities I
                ytype = SignalType.INTENSITY, ("I",)
                cols["I"] = cI, Column(None, None, SignalType.INTENSITY)
            case [AxisType.ENERGY, None, None, _, None, None, None]:
                # --intensities I0 I1
                raise InputFileParsingException(
                    "No y signal mode specified. Specify either transmission, fluorescence or intensity."
                )
            case [AxisType.ENERGY, [], None, [str(cI0), str(cI1)], None, None, None]:
                # --transmission --intensities I0 I1
                ytype = SignalType.TRANSMITTANCE, ("I0", "I1")
                cols["I0"] = cI0, Column(None, None, SignalType.INTENSITY)
                cols["I1"] = cI1, Column(None, None, SignalType.INTENSITY)
            case [AxisType.ENERGY, None, [], [str(cI0), str(cI1)], None, None, None]:
                # --fluorescence --intensities I0 I1
                raise InputFileParsingException(
                    "Fluorescence intensity (If) is not specified."
                )
            case [AxisType.ENERGY, [], None, [str(cI0), str(cI1), str(cI2)], None, None, None]:
                # --transmission --intensities I0 I1 I2
                ytype = SignalType.TRANSMITTANCE, ("I0", "I1")
                cols["I0"] = cI0, Column(None, None, SignalType.INTENSITY)
                cols["I1"] = cI1, Column(None, None, SignalType.INTENSITY)
                cols["I2"] = cI2, Column(None, None, SignalType.INTENSITY)
            case [AxisType.ENERGY, None, [], [str(cI0), str(cI1), str(cI2)], None, None, None]:
                # --fluorescence --intensities I0 I1 I2
                raise InputFileParsingException(
                    "Fluorescence intensity (If) is not specified."
                )
            case [AxisType.ENERGY,[],None,[str(cI0), str(cI1), str(cI2), str(cIf)],None, None, None]:
                # --transmission --intensities I0 I1 I2 If
                ytype = SignalType.TRANSMITTANCE, ("I0", "I1")
                cols["I0"] = cI0, Column(None, None, SignalType.INTENSITY)
                cols["I1"] = cI1, Column(None, None, SignalType.INTENSITY)
                cols["I2"] = cI2, Column(None, None, SignalType.INTENSITY)
                cols["If"] = cIf, Column(None, None, SignalType.INTENSITY)
            case [AxisType.ENERGY,None,[],[str(cI0), str(cI1), str(cI2), str(cIf)],None, None, None]:
                # --fluorescence --intensities I0 I1 I2 If
                ytype = SignalType.FLUORESCENCE, ("If", "I0")
                cols["I0"] = cI0, Column(None, None, SignalType.INTENSITY)
                cols["I1"] = cI1, Column(None, None, SignalType.INTENSITY)
                cols["I2"] = cI2, Column(None, None, SignalType.INTENSITY)
                cols["If"] = cIf, Column(None, None, SignalType.INTENSITY)
            case [AxisType.ENERGY, None, None, None, str(cI), None, None]:
                # --intensity I
                ytype = SignalType.INTENSITY, (cI,)
                cols["I"] = cI, Column(None, None, SignalType.INTENSITY)
            case [AxisType.ENERGY | AxisType.RELENERGY | AxisType.KVECTOR, None, None, None, None, str(cMu), None]:
                ytype = None
                cols["mu"] = cMu, Column(None, None, DataColType.MU)
            case [AxisType.ENERGY | AxisType.RELENERGY | AxisType.KVECTOR, None, None, None, None, None, str(cX)]:
                ytype = None
                cols["x"] = cX, Column(None, None, DataColType.CHI)
            case [AxisType.ENERGY, _, _, _, _, _, _]:
                raise InputFileParsingException("Too many signal specifications.")
            case _:
                raise InputFileParsingException(
                    "Only energy x-axis supports this type of data input."
                )

        match xaxis, args.reftransmittance, args.refabsorption:
            case [AxisType.ENERGY, None, None]:
                rtype = None
            case [AxisType.ENERGY, [], None]:
                if "I1" not in cols:
                    raise InputFileParsingException("No I1 column specified.")
                if "I2" not in cols:
                    raise InputFileParsingException("No I2 column specified.")
                rtype = SignalType.TRANSMITTANCE, ("I1", "I2")
            case [AxisType.ENERGY, [str(rI1), str(rI2)], None]:
                cols["rI1"] = rI1, Column(None, None, SignalType.INTENSITY)
                cols["rI2"] = rI2, Column(None, None, SignalType.INTENSITY)
                rtype = SignalType.TRANSMITTANCE, ("rI1", "rI2")
            case [AxisType.ENERGY, None, []]:
                raise InputFileParsingException("No reference signal mode specified.")
            case [AxisType.ENERGY, None, str(rI)]:
                cols["Ir"] = rI, Column(None, None, SignalType.INTENSITY)
                rtype = SignalType.INTENSITY, ("Ir",)
            case [AxisType.ENERGY, _, _]:
                raise InputFileParsingException(
                    "Too many reference signal specifications."
                )
            case _:
                raise InputFileParsingException(
                    "Only energy x-axis supports this type of data input."
                )

        if args.var is not None:
            vars = {
                str(k): attempt_parse_number(v) if str(v).isdecimal() else str(v)
                for k, v in args.var
            }
        else:
            vars = {}

        return Args_FileIn(
            args.filepos,
            args.dir,
            xaxis,
            xaxiscol,
            args.shift,
            cols,
            ytype, # type: ignore
            rtype,
            vars,
        )

    @staticmethod
    def execute(args: Args_FileIn, context: Context) -> CommandResult:
        log = getLogger("filein")

        # Check if the path is relative or absolute.
        relativeto = (
            Path(args.directory)
            if args.directory is not None
            else context.paths.workdir
        )
        files = get_filepaths(args.filepos, relativeto)
        # if os.path.isabs(args.filepos):
        #     _p = Path(args.filepos)
        #     _d = Path((_p.drive) + "/")
        #     _r = _p.relative_to(_d)
        #     files = [*_d.glob(_r.name)]
        #     # if _p.parent.name == "**":
        #     #     files = [*_p.parent.parent.rglob(_p.name)]
        #     # else:
        #     #     files = [*_p.parent.glob(_p.name)]
        # else:
        #     files = [*relativeto.glob(args.filepos)]

        if not files:
            raise FileNotFoundError(
                f"The specified file does not exist: {args.filepos}"
            )

        filecount = len(context.data.data)
        for i, file in enumerate(files, 1):
            data = read_file(file, args)
            context.data.add_data(data)

            data.meta.vars[".n"] = filecount + i
            data.meta.vars[".i"] = i
            log.info(f"Imported {data.meta.name}")

        return CommandResult(True)

    @staticmethod
    def undo(args: Args_FileIn, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError

class TestType(Enum):
    ENERGYCOLUMN = "energycolumn"
    GROUPS = "groups"

class Args_Test(NamedTuple):
    filepos: str
    testtype:TestType
    columns:str|None=None

class FileTest(CommandHandler):
    @staticmethod
    def parse(tokens: list[Token], context: Context) -> Args_Test:
        parser = CommandParser("test", description="Tests the file columns.")
        parser.add_argument("filename", help="The file to use for the test.")
        
        subparsers = parser.add_subparsers(dest="method")
        subparsers.add_parser("energy")
        subparsers.add_parser("groups")


        args = parser.parse(tokens)

        match args.method:
            case "energy":
                return Args_Test(args.filename, TestType.ENERGYCOLUMN, None)
            case "groups":
                return Args_Test(args.filename, TestType.GROUPS, None)
            case t: raise ValueError(f"Unknown test type: {t}.")

    @staticmethod
    def execute(args: Args_Test, context: Context) -> CommandResult:
        log = getLogger("test")

        file = get_filepaths(args.filepos, context.paths.workdir)[0]
        res, filedat, mdat = _read_file(file)

        match args.testtype:
            case TestType.ENERGYCOLUMN:
                # Calculate the difference between successive points. Energy
                # columns are strictly monotonic, so should be 0.
                diff = (filedat.diff() <= 0).sum()
                
            case TestType.GROUPS:
                pass

        return CommandResult(True)

    @staticmethod
    def undo(args: NamedTuple, context: Context) -> CommandResult:
        # TODO:
        raise NotImplementedError



from __future__ import annotations

import pandas as pd
import numpy as np
import numpy.typing as npt

from matplotlib.figure import Figure
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from pathlib import Path
from typing import Any, NamedTuple
from enum import Enum
from dataclasses import dataclass, field
from itertools import chain

from ._numberunit import Domain, NumberUnitRange, Bound, NumberUnit

@dataclass(slots=True, frozen=True)
class CommandResult:
    success: bool | None
    result: Any | None = None
    error: Any | None = None
    warning: Any | None = None


class Options(NamedTuple):
    interactive: bool
    verbose: bool
    version: tuple[int, ...]
    logcolor: bool
    updateinput: bool
    cache: bool
    other: dict[str, Any]


class Paths(NamedTuple):
    inputfile: Path | None
    currentdir: Path
    workdir: Path
    outputdir: Path
    userdir: Path
    logfile: Path
    configfile: Path


class AxisType(Enum):
    ENERGY = "eV"
    RELENERGY = "+eV"
    KVECTOR = "k"
    DISTANCE = "A"
    QVECTOR = "q"
    INDEX = "i"


class SignalType(Enum):
    FLUORESCENCE = "fl"
    TRANSMITTANCE = "tr"
    INTENSITY = "i"
    OTHER = "o"


class FourierType(Enum):
    COMPLEX = "c"
    PHASE = "p"
    ABSOLUTE = "a"
    REALPART = "r"
    IMAGPART = "i"


class DataColType(Enum):
    ALPHA = "a"
    CHI = "x"
    MU = "mu"
    PREEDGE = "pre"
    POSTEDGE = "post"
    BACKGROUND = "bkg"
    WINDOW = "win"
    MULTIEDGE = "mult"
    GLITCH = "gli"


class FigureSettings(NamedTuple):
    figurenum: int
    subplot: tuple[int,int]

@dataclass(slots=True)
class AxisRuntime:
    axis: Axes
    xlimits: tuple[NumberUnit | Bound | None, NumberUnit | Bound | None] = None, None
    ylimits: tuple[NumberUnit | Bound | None, NumberUnit | Bound | None] = None, None
    _lines: list[tuple[npt.NDArray, npt.NDArray]] = field(default_factory=list)


@dataclass(slots=True)
class FigureRuntime:
    settings: FigureSettings
    figure: Figure
    axes: dict[tuple[int,int], AxisRuntime]
    shown: bool = False

    def show(self) -> None:
        def on_close(event): self.figure.canvas.stop_event_loop()
        self.figure.canvas.mpl_connect('close_event', on_close)
        self.figure.show()
        self.figure.canvas.start_event_loop(timeout=-1)
        self.shown = True
    
    @classmethod
    def new(cls, settings:FigureSettings) -> FigureRuntime:
        _fig = plt.figure(settings.figurenum)

        match settings.subplot:
            case (1,1):
                _ax = {(1,1):AxisRuntime(_fig.subplots(1,1))}
            case (1,_):
                _axs = _fig.subplots(*settings.subplot)
                _ax = {(1,c):AxisRuntime(_ax) for c,_ax in enumerate(_axs,1)}
            case (_,1):
                _axs = _fig.subplots(*settings.subplot)
                _ax = {(r,1):AxisRuntime(_ax) for r,_ax in enumerate(_axs,1)}
            case (_,_):
                _axss = _fig.subplots(*settings.subplot)
                _ax = {(r,c):AxisRuntime(_ax) for r,_axs in enumerate(_axss,1) for c,_ax in enumerate(_axs,1)}
            case _: raise RuntimeError("Unknown error: #3409234")
            

        figure = FigureRuntime(settings, _fig, _ax, False)
        return figure

class Figures(NamedTuple):
    impl_figsettings: list[FigureSettings]
    expl_figsettings: dict[int, FigureSettings]
    figureruntimes: dict[int, FigureRuntime]

    def get_impl_figurenum(self) -> int:
        # To be used for figures created without --fig flag
        allfignums = chain(self.expl_figsettings, self.figureruntimes)
        usrfignums = filter(lambda f:f <= 1000, allfignums)
        return max(usrfignums, default=1) +1

    def get_high_figurenum(self) -> int:
        # To be used for figures created by commands. Figures will have fignum > 1000
        allfignums = chain(self.expl_figsettings, self.figureruntimes)
        usrfignums = filter(lambda f:f > 1000, allfignums)
        return max(usrfignums, default=1000)


ColumnType = AxisType | SignalType | DataColType | FourierType


@dataclass(slots=True)
class MetaData:
    signaltype: SignalType | None
    refsigtype: SignalType | None
    name: str
    path: Path
    vars: dict[str, str | int | float]
    run: dict[str, Any] = field(default_factory=dict)
    refE0: float | None = None
    E0: float | None = None

    def __contains__(self, varname:str) -> bool:
        match varname:
            case ".st" if self.signaltype is not None: return True
            case ".rt" if self.refsigtype is not None: return True
            case "E0" if self.E0 is not None: return True
            case "rE0" if self.refE0 is not None: return True
            case var if var in self.vars: return True
        return False


    def get(self, varname:str, *, default:Any=...) -> str | int | float:
        if varname not in self:
            if default is not ...:
                return default
            raise RuntimeError(f"Variable {varname} is not defined.")
        match varname:
            case ".st" if self.signaltype is not None:
                return self.signaltype.value
            case ".rt" if self.refsigtype is not None:
                return self.refsigtype.value
            case "E0" if self.E0 is not None:
                return self.E0
            case "rE0" if self.refE0 is not None:
                return self.refE0
            case var if var in self.vars:
                return self.vars[var]
            case var:
                raise RuntimeError(f"Variable {var} is not defined.")

@dataclass(slots=True, frozen=True)
class Column:
    unit: str | None
    sign: bool | None
    axis: ColumnType | None

class Datum:
    __slots__ = "df","cols","default_axis"
    # df: pd.DataFrame
    # cols: dict[str, Column]
    # default_axis: str | None

    def __init__(self, df:pd.DataFrame|None=None, cols:dict[str,Column]|None=None) -> None:
        self.default_axis:str|None = None
        match df,cols:
            case None,None:
                self.df = pd.DataFrame()
                self.cols:dict[str, Column] = {}
            case df,cols if df is not None and cols is not None:
                self.df = df
                self.cols = cols
            case _: raise AttributeError("Cannot instanciate if only one of df, cols is None")

    def set_default_axis(self, colname:str) -> None:
        if colname not in self.cols: raise KeyError(f"Column {colname} does not exist.")
        if not isinstance(self.cols[colname].axis, AxisType):
            raise ValueError(f"Column {colname} is not an axis column.")
        self.default_axis = colname

    def get_default_axis(self) -> pd.Series:
        if self.default_axis is None: raise KeyError("No default axis specified.")
        return self.df[self.default_axis]

    def add_col(self, values:npt.NDArray[np.floating | np.complexfloating] | pd.Series, coltype: Column, colname: str, *, unique:bool=True) -> str:
        if colname in self.cols:
            if unique:
                raise NameError("A column with the given name already exists.")
            else:
                n = len([c for c in self.cols if c.startswith(colname) and not c.startswith(f"{colname}_")])
                colname = f"{colname}{n}"

        self.df.loc[:,colname] = values
        self.cols[colname] = coltype
        return colname

    def mod_col(self, colname:str, values:npt.NDArray | pd.Series) -> None:
        if colname not in self.cols: raise KeyError(f"No column with the given name: {colname}")
        # Valid column names are "name" (active column) or "name_0", "name_1", ... (old modified column)
        n = max([int(c.split("_", maxsplit=1)[1]) for c in self.cols if c.startswith(f"{colname}_")], default=-1)+1

        old = f"{colname}_{n}"
        self.df.rename(columns={colname:old}, inplace=True)
        self.df.loc[:,colname] = values

        self.cols[old] = self.cols[colname]
    
    def get_cols(self, colname:str|None=..., *, coltype: ColumnType=...) -> pd.DataFrame: # type: ignore
        if colname is None: return self.df.index.to_frame()
        if colname is not ...: return self.df.loc[:,[colname]]
        if coltype is not ...: return self.df.loc[:,[c for c,ct in self.cols.items() if ct.axis==coltype]]
        return self.df

class Data:
    def __init__(self, metadata: MetaData) -> None:
        self.datums: dict[Domain, Datum] = {}
        # meta holds the metadata
        self.meta = metadata
    
    def add_col(self, colname:str, values:npt.NDArray | pd.Series, column: Column, domain:Domain) -> None:
        if domain not in self.datums:
            self.datums[domain] = Datum()
        
        self.datums[domain].add_col(values, column, colname)
    
    def _get_col_domain(self, colname:str) -> Domain:
        _ds = [domain for domain, datum in self.datums.items() if colname in datum.cols]
        if len(_ds) == 0: raise KeyError(f"column {colname} was not found")
        elif len(_ds) == 1: return _ds[0]
        else: raise KeyError("Multiple matching columns.")

    def _get_all_cols(self) -> list[tuple[str, Column, Domain]]:
        return [(colname, coltype, domain)
                for domain,datum in self.datums.items()
                for colname,coltype in datum.cols.items()]

    def mod_col(self, colname:str, values:npt.NDArray | pd.Series) -> None:
        domain = self._get_col_domain(colname)
        self.datums[domain].mod_col(colname, values)

    def get_cols(self, colname:str|None = ..., *, coltype: ColumnType = ..., domain: Domain = ...) -> pd.DataFrame: # type: ignore
        if domain is not ...:
            return self.datums[domain].get_cols(colname, coltype=coltype)
        for domain,datum in self.datums.items():
            try:
                r = datum.get_cols(colname, coltype=coltype)
            except KeyError: continue
            if len(r.columns): return r
        return pd.DataFrame()

    def get_col(self, colname:str|None = ..., *, coltype: ColumnType = ..., domain: Domain = ...) -> pd.Series: # type: ignore
        cols = self.get_cols(colname, coltype=coltype, domain=domain)
        if len(cols.columns) > 1: raise ValueError("More than one columns match.")
        return cols[cols.columns[0]]
    def get_col_(self, colname:str|None = ..., *, coltype: ColumnType = ..., domain: Domain = ...) -> npt.NDArray: # type: ignore
        return self.get_col(colname, coltype=coltype, domain=domain).to_numpy()
    
    def get_xy_(self, xname:str, yname:str) -> tuple[npt.NDArray, npt.NDArray]:
        xd = self._get_col_domain(xname)
        yd = self._get_col_domain(yname)
        if xd != yd: raise KeyError("X and Y columns are on different domains.")

        return self.get_col_(xname, domain = xd), self.get_col_(yname, domain = xd)

    def get_xy(self, xname:str, yname:str) -> pd.Series:
        x,y = self.get_xy_(xname, yname)
        return pd.Series(y, x, name=yname)


class DataStore:
    def __init__(self) -> None:
        self.data: list[Data] = []

    def add_data(self, data: Data) -> None:
        self.data.append(data)

    def __iter__(self):
        return self.data.__iter__()


class Directives(NamedTuple):
    clear: bool
    noplot: bool
    vars: dict[str, str]


class Context:
    def __init__(
        self,
        paths: Paths,
        config: dict[str, Any],
        options: Options,
        directives: Directives,
    ) -> None:
        self.paths = paths
        self.config = config
        self.options = options
        self.directives = directives

        self.vars: dict[str, Any] = {}
        self.commands: list[tuple[NamedTuple, CommandResult]] = []
        self.data = DataStore()
        self.figures = Figures([], {},{})


def range_to_index(data: Data, range: NumberUnitRange) -> pd.Series:
    if range.domain is None: raise KeyError("Unknown index domain.")
    if range.inter is not None:
        raise ValueError("Index selection cannot be done if interval is not None.")
    if isinstance(range.lower, Bound) or isinstance(range.upper, Bound):
        raise ValueError("Index selection must be done only if the bounds are actualized.")
    
    datum = data.datums[range.domain]
    # Get only the live columns
    cols = [col for col in datum.cols if len(col.split("_")) == 1]
    
    for colname in cols:
        col = datum.cols[colname]
        if range.lower.unit != col.unit: continue
        match col.sign, range.lower.sign:
            case [None, _]:...
            case [False, 0]:...
            case [True, -1|1]:...
            case _: continue
        idx_l = datum.df[colname] >= range.lower.value
        break
    else:
        raise KeyError("No matching column found.")
    
    
    for colname in cols:
        col = datum.cols[colname]
        if range.upper.unit != col.unit: continue
        match col.sign, range.upper.sign:
            case [None, _]:...
            case [False, 0]:...
            case [True, -1|1]:...
            case _: continue
        idx_u = datum.df[colname] <= range.upper.value
        break
    else:
        raise KeyError("No matching column found.")
    
    return idx_l&idx_u

def range_to_index_(data: Data, range: NumberUnitRange) -> npt.NDArray[np.bool_]:
    return range_to_index(data, range).to_numpy()